#include <cstdio>
#include <iostream>
#include <string>
#include <cstring>
#include <cmath>
#include <vector>
#include <queue>
#include <cstdlib>
#include <ctime>
#include <algorithm>
#include <sstream>
#include <map>
#include <random>
#include <ctime>
#include <iterator>
#include <numeric>

using namespace std;

default_random_engine generator;

using Point = vector< double >;
using Points = vector< Point >;

vector<int> labels;

Points points;

const double epsi = 1e-9;
const double oo = 1e100;

int n = 0;
int d = 0;

void generate_sythetic_data(int n, int d, int num_clusters)
{
	int n_per_cluster = n / num_clusters;
	
	points.resize(n);
	labels.resize(n);
	
	clock_t time_length = clock();
	cerr << "*******************************************\n";
	cerr << "n " << n << " d " << d << " k " << num_clusters << endl;
	cerr << "*******************************************\n";
	
	uniform_real_distribution<double> uniform_distribution_1(-50, 50);
	
	Points centers(num_clusters);
	
	for(int i = 0; i < num_clusters; i++)
	{
		centers[i].resize(d);
		for(int j = 0; j < d; j++)
			centers[i][j] = uniform_distribution_1(generator);
	}
	
	for(int i = 0; i < num_clusters; i++)
	{
		normal_distribution<double> guass_distribution(0, 1);
		for(int j = 0; j < n_per_cluster; j++)
		{
			int x = i * n_per_cluster + j;
			labels[x] = i;
			points[x].resize(d);
			for(int k = 0; k < d; k++)
				points[x][k] = centers[i][k] + guass_distribution(generator);
		}
	}
	
	time_length = clock() - time_length;
	double seconds_needed = (double)time_length / CLOCKS_PER_SEC;
	cerr << "*******************************************\n";
	cerr << "Generated! Generating time " << seconds_needed << endl;
	cerr << "*******************************************\n";
}

inline double sq_dis(const Point& a, const Point &b)
{
	double ans = 0;
	
	for(int i = 0; i < d; i++)
		ans += (a[i] - b[i]) * (a[i] - b[i]);
		
	return ans;
}

Points kcenters;
vector<int> clusters_sizes;

double get_kmeans_cost(const vector<int>& cluster_label)
{
	double total_cost = 0;
	
	clusters_sizes.resize(n);
	kcenters.resize(n);
	for(int i = 0; i < n; i++)
	{
		kcenters[i].resize(d);
		for(int j = 0; j < d; j++)
			kcenters[i][j] = 0;
		clusters_sizes[i] = 0;
	}
	
	for(int i = 0; i < n; i++)
		clusters_sizes[cluster_label[i]]++;
		
	for(int i = 0; i < n; i++)
	{
		for(int j = 0; j < d; j++)
			kcenters[cluster_label[i]][j] += points[i][j] / clusters_sizes[cluster_label[i]];
	}
	
	for(int i = 0; i < n; i++)
		total_cost += sq_dis(points[i], kcenters[cluster_label[i]]);
	
	return total_cost;
}

int main()
{
	vector<pair<int,int>> candidates_n_d = {pair<int,int>(100000,10),pair<int,int>(100000,100)};
	vector<int> candidates_k = {5, 10, 25, 100, 1000};
	double true_cost = 0;
	
	int i = 0;
	
	for(pair<int,int> s : candidates_n_d )
	{
		n = s.first, d = s.second;
		for(int k : candidates_k)
		{
			generate_sythetic_data(n,d,k);
			true_cost = get_kmeans_cost(labels);
			cout<<"k-means cost: "<<true_cost<<endl;
			
			char str[25] = {0};
			sprintf(str, "%d", i);
			string dataset_name =  (string)"input" + str + ".txt";
			string label_name = (string)"label" + str + ".txt";
			i++;
			
			FILE *fp = fopen(dataset_name.c_str(), "w");
			for(int p = 0; p < n; p++)
				for(int q = 0; q < d; q++)
					fprintf(fp, "%.3lf%c", points[p][q],q==d-1?'\n':' ');
			fclose(fp);
			
			fp = fopen(label_name.c_str(), "w");
			for(int p = 0; p < n; p++)
				fprintf(fp,"%d\n", labels[p]);
			fclose(fp);
		}
	}
	
	
	return 0;
}

