from sklearn.datasets import fetch_openml, fetch_rcv1
from sklearn import preprocessing
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA, KernelPCA
from sklearn.cluster import KMeans
from sklearn.utils import Bunch

import numpy as np
from copy import deepcopy


# K-means
def kmeans(X, k):
    alg = KMeans(init='k-means++', n_clusters=k, n_init=10)
    alg.fit(X)
    return alg

def read_data(str):
  str_data = str + '/input.txt'
  str_label = str + '/label.txt'
  A = np.loadtxt(str_data)
  labels = np.loadtxt(str_label)
  labels = labels - np.min(labels)
  num_classes = np.max(labels) + 1
  return A, labels.astype(int), num_classes.astype(int)

import matplotlib.pyplot as plt
from sklearn.metrics.cluster import adjusted_mutual_info_score
from sklearn.metrics.cluster import adjusted_rand_score
plt.style.use('seaborn-whitegrid')
def dis(x, y):
  return np.sqrt(np.sum((x-y)*(x-y)))

def dis_to_centers(A, labels, centers):
  n = np.size(A, 0)
  d = np.size(A, 1)
  num_classes = np.max(labels) + 1
  exist_labels = np.zeros(num_classes)
  for i in range(n):
    exist_labels[labels[i]] = 1
  dis_to_center = np.zeros(n);
  dis_to_off_center = np.zeros(n);
  ratio = np.zeros(n)
  for i in range(n):
    dis_to_center[i] = dis(A[i], centers[labels[i]]);
    dis_to_off_center[i] = np.Inf
    for j in range(num_classes):
      if(j != labels[i] and dis(A[i], centers[j]) < dis_to_off_center[i] and exist_labels[j] > 0.1):
          dis_to_off_center[i] = dis(A[i], centers[j])
    ratio[i] = dis_to_off_center[i] / (dis_to_center[i] + 1e-9)
  
  return dis_to_center, dis_to_off_center, ratio
  
def show_ratio(ratio, name):
  fig = plt.figure()
  ax = plt.axes()
  ax.set_xlabel('fraction of points')
  ax.set_ylabel('outer-dis / inner-dis')
  plt.ylim(0,10)
  plt.yticks(np.arange(1, 10, 1))
  plt.xticks(np.arange(0, 1, 0.1))
  fig.suptitle(name)

  x = np.linspace(0, 1, np.size(ratio))
  pos = np.argsort(-ratio)
  n=np.size(pos)
  print("last:",ratio[pos[n-1]],"0.7:",ratio[pos[int(0.7*n)]])
  ax.plot(x, ratio[pos])
  plt.show()

def print_data_info(data, labels, ground_truth_k):
  print("data size:", np.size(data, 0), np.size(data, 1))
  print("data labels:", np.max(labels)-np.min(labels)+1)
  print("ground_truth_k:", ground_truth_k)

  res = kmeans(data, ground_truth_k)
  print("clustering_cost:", res.inertia_)
  
  dis_to_center, dis_to_off_center, ratio = dis_to_centers(data, res.labels_, res.cluster_centers_)
  show_ratio(ratio, "dataset")
  
dataset_names = ['EUCA', 'MICE', 'BIOMED', 'DIABETES', 'MAGIC', 'GESTURE', 'mnist', 'zoo']
for name in dataset_names:
  print(name)
  data, labels, ground_truth_k = read_data("dataset/"+name) 
  print_data_info(data, labels, ground_truth_k)
  

