#include <cstdio>
#include <iostream>
#include <string>
#include <cstring>
#include <cmath>
#include <vector>
#include <queue>
#include <cstdlib>
#include <ctime>
#include <algorithm>
#include <sstream>
#include <map>
#include <random>
#include <ctime>
#include <iterator>
#include <numeric>

using namespace std;

////////////////////////////////////////////////////////////////////////////////////////// EXPERIMENT CONFIGS

const char name_str[25][205] = {
			"dataset/synthetic/input0.txt",
			"dataset/synthetic/input1.txt",
			"dataset/synthetic/input2.txt",
			"dataset/synthetic/input3.txt",
			"dataset/synthetic/input4.txt",
			"dataset/synthetic/input5.txt",
			"dataset/synthetic/input6.txt",
			"dataset/synthetic/input7.txt",
			"dataset/synthetic/input8.txt",
			"dataset/synthetic/input9.txt"
			};
			

const int target_k[25] ={
5,
10,
25,
100,
1000,
5,
10,
25,
100,
1000
};


////////////////////////////////////////////////////////////////////////////////////////// UTILS

using Point = vector< double >;
using Points = vector< Point >;
default_random_engine generator;

Points points;

const double epsi = 1e-9;
const double oo = 1e100;

int n = 0;
int d = 0;

inline bool equal(const Point&a, const Point &b)
{
	for(int i =0; i < a.size(); i++)
		if(a[i] < b[i] - epsi || b[i] < a[i] - epsi)
			return false;
	return true;
}

inline double dis(const Point& a, const Point &b)
{
	double ans = 0;
	
	for(int i = 0; i < d; i++)
		ans += (a[i] - b[i]) * (a[i] - b[i]);
		
	return sqrt(ans); 
}

inline double sq_dis(const Point& a, const Point &b)
{
	double ans = 0;
	
	for(int i = 0; i < d; i++)
		ans += (a[i] - b[i]) * (a[i] - b[i]);
		
	return ans;
}

Points kcenters;
vector<int> clusters_sizes;

double get_kmeans_cost(const vector<int>& cluster_label)
{
	double total_cost = 0;
	
	clusters_sizes.resize(n);
	kcenters.resize(n);
	for(int i = 0; i < n; i++)
	{
		kcenters[i].resize(d);
		for(int j = 0; j < d; j++)
			kcenters[i][j] = 0;
		clusters_sizes[i] = 0;
	}
	
	for(int i = 0; i < n; i++)
		clusters_sizes[cluster_label[i]]++;
		
	for(int i = 0; i < n; i++)
	{
		for(int j = 0; j < d; j++)
			kcenters[cluster_label[i]][j] += points[i][j] / clusters_sizes[cluster_label[i]];
	}
	
	for(int i = 0; i < n; i++)
		total_cost += sq_dis(points[i], kcenters[cluster_label[i]]);
	
	return total_cost;
}



////////////////////////////////////////////////////////////////////////////////////////// READ DATA 

void read_data(const string &data_name)
{
	const int buffer_size = 10000000;
	static char buf[buffer_size] = {0};

	n = d = 0;
	points.clear();
	
	string points_name = data_name;
	
	FILE *fp = fopen(points_name.c_str(), "r");
	
	for(;fgets(buf, buffer_size - 1, fp) != NULL; n++)
	{
		stringstream sin(buf);
		Point x;
		
		if(d == 0)
		{
			int i = 0;
			double y;
			for(; sin >> y; i++) x.push_back(y);
			d = i;
		}
		else
		{
			x.resize(d);
			for(int i = 0; sin >> x[i]; i++);
		}
		
		points.push_back(x);
	}
		
	cerr << "*******************************************\n";
	cerr << "n " << n << " d " << d << endl;
	cerr << "n " << points.size() << " d " << points[0].size() << endl;
	cerr << "*******************************************\n";
	
	fflush(stdout);
	fclose(fp);
}

////////////////////////////////////////////////////////////////////////////////////////// UNION FIND FOR CONNECTED COMPONENT 

int find(int x, vector<int> &father)
{
	if(father[x] == x)
		return father[x];
	return father[x] = find(father[x], father);
}

////////////////////////////////////////////////////////////////////////////////////////// APPLY LSH

Points lsh_points;

void apply_lsh(double scale)
{
	double r = 4;
	int hash_d = 10;
	Points guass_points(hash_d);
	vector<double> shift_vec(hash_d);
	
	normal_distribution<double> guass_distribution(0, 1);
	uniform_real_distribution<double> uniform_distribution(0.0, r);
	
	for(int i = 0; i < hash_d ;i++)
	{
		guass_points[i].resize(d);
		for(int j = 0; j < d; j++)
		{
			guass_points[i][j] = guass_distribution(generator)/scale;
			//cerr<<guass_points[i][j]<<' ';
		}
		shift_vec[i]=uniform_distribution(generator);
		//cerr<<shift_vec[i]<<endl;
	}
	
	lsh_points.resize(n);
	for(int i = 0; i < n; i++)
	{
		lsh_points[i].resize(hash_d);
		for(int j = 0; j < hash_d; j++)
		{
			lsh_points[i][j] = 0;
			for(int k = 0; k < d ;k++)
				lsh_points[i][j] += points[i][k] * guass_points[j][k];
			lsh_points[i][j] += shift_vec[j];
			lsh_points[i][j] = floor(lsh_points[i][j]/r);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////// BUILD NEAR NEIGHBOR GRAPH

vector<int> connects;
vector<int> idx;

bool cmp(int p, int q)
{
	return lsh_points[p] < lsh_points[q];
}

void near_neighbor_graph(double scale)
{
	int t = 100;
	double max_dis = 0;
	
	connects.resize(n);
	idx.resize(n);
	for(int i = 0; i < n; i++)
		connects[i] = i;
	
	for(; t; t--)
	{
		apply_lsh(scale);
		for(int i = 0; i < n; i++)
			idx[i] = i;
		sort(idx.begin(),idx.end(),cmp);
		int last = idx[0];
		for(int i = 0; i < n - 1; i++)
		{
			int p = idx[i], q = idx[i + 1];
			if(equal(lsh_points[p],lsh_points[q]))
			{
				if(dis(points[q], points[last]) <= scale)
				{
					connects[find(q,connects)] = find(last, connects);
					
					if(dis(points[q], points[last]) > max_dis)
						max_dis = dis(points[q], points[last]);
				}
			}
			else
				last = q;
		}
	}
	
	for(int i = 0; i < n; i++)
		find(i, connects);
		
	/*for(int i = 0; i < n; i++)
		for(int j = 0; j < n; j++)
		{
			if(connects[i] != connects[j] && dis(points[i], points[j]) < min_dis)
				min_dis = dis(points[i], points[j]);
			if(connects[i] == connects[j] && dis(points[i], points[j]) > max_real_dis)
				max_real_dis = dis(points[i], points[j]);
		}
	
	cerr<<"scale: "<<scale<<" max dis: "<<max_dis<<" min dis: "<<min_dis<<" max real dis: "<<max_real_dis<<endl;*/
	cerr<<"scale: "<<scale<<" max dis: "<<max_dis<<endl;
}

////////////////////////////////////////////////////////////////////////////////////////// FIND CANDIDATE CLUSTERS

vector<vector<int>> candidates;
vector<vector<double>> W;
int L;

vector<int> father;
vector<int> last_father;
vector<double> weight;
Points centers;
vector<int> num_points; 

void candidate_clusters()
{
	double scale = 0.01;
	double max_dis = 0;
	
	cerr<<"##############################start to build clusters#########################"<<endl;
	candidates.clear();
	W.clear();
	L = 0;
	
	last_father.resize(n);
	father.resize(n);
	weight.resize(n);
	centers.resize(n);
	num_points.resize(n);
	
	for(int i = 0; i < n; i++)
	{
		father[i] = i;
		weight[i] = 0;
		centers[i].resize(d);
		for(int j = 0; j < d; j++)
			centers[i][j] = 0;
		num_points[i] = 0;
	}
	
	candidates.push_back(father);
	last_father = father;
	W.push_back(weight);
	cerr<< L <<" num clusters:"<< n <<endl;
	L++;
	
	for(int i = 0; i < n;i++)
	{
		double tmp = dis(points[i], points[0]);
		if(tmp > max_dis) max_dis = tmp;
	}
	
	for(;scale <= 2 * max_dis; scale*=1.1)
	{
		near_neighbor_graph(scale);
		for(int i = 0; i < n; i++)
		{
			int x = i;
			int y = connects[i];
			father[find(x, father)] = father[find(y, father)];
			
			centers[i].resize(d);
			for(int j = 0; j < d; j++)
				centers[i][j] = 0;
			num_points[i] = 0;
			weight[i] = 0;
		}
		for(int i = 0; i < n; i++)
		{
			find(i, father);
			num_points[father[i]]++;
			for(int j = 0; j < d; j++)
				centers[father[i]][j] += points[i][j];
		}
		for(int i = 0; i < n; i++)
			if(father[i] == i)
				for(int j = 0; j < d; j++)
					centers[i][j] /= num_points[i];
		for(int i = 0; i < n; i++)
			weight[father[i]] += sq_dis(points[i], centers[father[i]]);
		if(father != last_father)
		{
			last_father = father;
			candidates.push_back(father);
			W.push_back(weight);
			int tot = 0;
			for(int i = 0; i < n; i++)
				if(i == father[i])
					tot++;
			cerr<< L <<" num clusters:"<< tot << endl;
			L++;
		}
	}
	
	for(int i = 0; i < n; i++)
	{
		father[i] = father[0];
		weight[i] = 0;
	}
	Point center(d);
	for(int i = 0; i < n; i++)
		for(int j = 0; j < d; j++)
			center[j] += points[i][j] / n;
	for(int i = 0; i < n; i++)
		weight[father[0]] += sq_dis(points[i], center);
	
	if(father != last_father)
	{
		last_father = father;
		candidates.push_back(father);
		W.push_back(weight);
		cerr<< L <<" num clusters:"<< 1 <<endl;
		L++;
	}
	
	cerr<<"##############################finish building clusters#########################"<<endl;
}

////////////////////////////////////////////////////////////////////////////////////////// BUILD BINARY TREE

vector<int> process_order;
vector<vector<int>> children;
vector<double> final_weights;
int level;
int root;

bool cmp2(int p, int q)
{
	return candidates[level][p] < candidates[level][q];
}

void build_binary_tree()
{
	process_order.clear();
	children.resize(L * n);
	final_weights.resize(2 * L * n);
	for(int i = 0; i < L * n; i++)
	{
		final_weights[i] = W[i / n][i % n];
		final_weights[i + L * n] = oo;
		children[i].resize(2);
		children[i][0] = children[i][1] = -1;
	}
	
	int o = L * n;
	
	for(int i = 0; i < L - 1; i++)
	{
		idx.clear();
		for(int j = 0; j < n; j++)
		{
			if(candidates[i][j] == j)
				idx.push_back(j);
			process_order.push_back(i * n + j);
		}
		
		level = i + 1;
		sort(idx.begin(),idx.end(),cmp2);
		
		for(int j = 0; j < idx.size(); )
		{
			int p = j;
			for(; p < idx.size() && candidates[i + 1][idx[j]]==candidates[i + 1][idx[p]]; p++);
			
			int s = p - j;
			for(int q = 0; q < s; q++)
			{
				vector<int> tmp;
				children.push_back(tmp);
				children[o + q].resize(2);
				children[o + q][0] = children[o + q][1] = -1;
				final_weights[o + q] = oo;
			}
			
			for(int q = j; q < p; q++)
			{
				int x = idx[q];
				int y = o + ((q - j + 1) + s) / 2 - 1;
				if(children[y][0] == -1)
					children[y][0] = i * n + x;
				else
					children[y][1] = i * n + x;
			}
			children[(i + 1) * n + candidates[i+1][idx[j]]][0] = o;
			for(int q = o + 1; q < o + s; q++)
			{
				int y = o + (q- o + 1)/2 - 1;
				if(children[y][0]==-1)
					children[y][0] = q;
				else
					children[y][1] = q;
			}
			for(int q = o + s - 1; q>=o; q--)
				process_order.push_back(q);
			
			o = o + s;
			j = p;
		}
	}
	process_order.push_back((L-1) * n + candidates[L-1][0]);
	root = (L-1) * n + candidates[L-1][0];
}

////////////////////////////////////////////////////////////////////////////////////////// DP BY COST

vector<vector<double>> DPcost;
vector<vector<int>> cost_best_strategy;

void dp_cost(int k)
{
	DPcost.resize(2 * L * n);
	cost_best_strategy.resize(2 * L * n);
	for(int i = 0; i < 2 * L * n; i++)
	{
		DPcost[i].resize(k + 1);
		cost_best_strategy[i].resize(k + 1);
		for(int j = 0; j <= k ;j++)
		{
			DPcost[i][j] = oo;
			cost_best_strategy[i][j] = -1;
		}
	}
	
	for(int i = 0; i < process_order.size(); i++)
	{
		int x = process_order[i];
		for(int j = 1; j <= k; j++)
		{
			DPcost[x][j] = final_weights[x];
			cost_best_strategy[x][j] = 0;
		}	
			
		if(children[x][0] != -1 && children[x][1] == -1)
		{
			int y = children[x][0];
			for(int j = 1; j <= k; j++)
			{
				if(DPcost[y][j] < DPcost[x][j])
				{
					DPcost[x][j] = DPcost[y][j];
					cost_best_strategy[x][j] = j;
				}
			}
		}
		else if(children[x][0] !=-1 && children[x][1] != -1)
		{
			int y1 = children[x][0];
			int y2 = children[x][1];
			
			for(int j = 1; j <= k; j++)
				for(int p = 1; p < j; p++)
				{
					if(DPcost[y1][p] + DPcost[y2][j-p] < DPcost[x][j])	
					{
						DPcost[x][j] = DPcost[y1][p] + DPcost[y2][j-p];
						cost_best_strategy[x][j] = p;
					}
				}
		}	
	}
}

vector<int> labels_by_cost;

void recurse_find_dp_cost_solution(int x, int i, vector<vector<bool>> &chosen)
{
	if(cost_best_strategy[x][i] == 0)
	{
		chosen[x / n][x % n] = true;
		return;
	}
	if(children[x][1] == -1)
	{
		recurse_find_dp_cost_solution(children[x][0], i, chosen);
	}
	else
	{
		recurse_find_dp_cost_solution(children[x][0], cost_best_strategy[x][i] , chosen);
		recurse_find_dp_cost_solution(children[x][1], i - cost_best_strategy[x][i] , chosen);
	}
}

double find_solution_cost(int k)
{
	vector<vector<bool>> chosen;
	
	labels_by_cost.resize(n);
	chosen.resize(L);
	for(int i = 0; i < L; i++)
	{
		chosen[i].resize(n);
		for(int j = 0; j < n; j++)
			chosen[i][j] = false;
	}
	for(int i = 0; i < n; i++)
		labels_by_cost[i] = -1;
	
	recurse_find_dp_cost_solution(root, k, chosen);
	
	for(int i = 0; i < L; i++)
		for(int j = 0; j < n; j++)
			if(chosen[i][candidates[i][j]])
				labels_by_cost[j] = candidates[i][j];
	
	return get_kmeans_cost(labels_by_cost);
}

////////////////////////////////////////////////////////////////////////////////////////// DP BY SIZE

vector<vector<double>> DPsize;
vector<vector<int>> size_best_strategy;
vector<double> lambdas;
vector<vector<int>> corresponds_i;
int l_size;

void dp_size(double scaler)
{
	double smallest = 0.1;
	double largest = final_weights[root];
	
	l_size = 0;
	lambdas.clear();
	lambdas.push_back(0);
	l_size++;
	lambdas.push_back(smallest);
	l_size++;
	for(;lambdas[l_size - 1] * scaler <= 2 * largest; l_size++)
		lambdas.push_back(lambdas[l_size - 1] * scaler);
	cerr<<"l_size: "<<l_size<<endl<<flush;


	corresponds_i.resize(l_size);
	for(int i = 0; i < l_size; i++)
	{
		corresponds_i[i].resize(l_size);
		for(int j = 0; j < l_size; j++)
		{
			corresponds_i[i][j] = l_size + 1;
			for(int p = 0; p < l_size; p++)
			{
				if(lambdas[i] >= lambdas[j] + lambdas[p])
					corresponds_i[i][j] = p;
			}
		}
	}
	
	DPsize.resize(2 * L * n);
	size_best_strategy.resize(2 * L * n);
	for(int i = 0; i < 2 * L * n; i++)
	{
		DPsize[i].resize(l_size);
		size_best_strategy[i].resize(l_size);
		for(int j = 0; j < l_size; j++)
		{
			DPsize[i][j] = n;
			size_best_strategy[i][j] = -1;
		}
	}
	
	for(int i = 0; i < process_order.size(); i++)
	{
		int x = process_order[i];
		for(int j = 0; j < l_size; j++)
		{
			if(final_weights[x]<=lambdas[j])
			{
				DPsize[x][j] = 1;
				size_best_strategy[x][j] = -1;
			}
		}
		
		if(children[x][0] != -1 && children[x][1] == -1)
		{
			int y = children[x][0];
			for(int j = 0; j < l_size; j++)
				if(DPsize[y][j] < DPsize[x][j])
				{
						DPsize[x][j] = DPsize[y][j];
						size_best_strategy[x][j] = j;
				}
		}
		else if(children[x][0] !=-1 && children[x][1] != -1)
		{
			int y1 = children[x][0];
			int y2 = children[x][1];
			
			for(int j = 0; j < l_size; j++)
				for(int p = 0; p <= j; p++)
				{
					int q = corresponds_i[j][p];
					if(DPsize[y1][p]+DPsize[y2][q] <= DPsize[x][j])
					{
						DPsize[x][j] = DPsize[y1][p]+DPsize[y2][q];
						size_best_strategy[x][j] = p;
					}
				}
		}
		
	}
}

vector<int> labels_by_size;

void recurse_find_dp_size_solution(int x, int i, vector<vector<bool>> &chosen)
{	
	if(size_best_strategy[x][i] == -1)
	{
		chosen[x / n][x % n] = true;
		return;
	}
	if(children[x][1] == -1)
	{
		recurse_find_dp_size_solution(children[x][0], i, chosen);
	}
	else
	{
		recurse_find_dp_size_solution(children[x][0], size_best_strategy[x][i] , chosen);
		recurse_find_dp_size_solution(children[x][1], corresponds_i[i][size_best_strategy[x][i]] , chosen);
	}
}

double find_solution_size(int k)
{
	vector<vector<bool>> chosen;
	
	labels_by_size.resize(n);
	chosen.resize(L);
	for(int i = 0; i < L; i++)
	{
		chosen[i].resize(n);
		for(int j = 0; j < n; j++)
			chosen[i][j] = false;
	}
	for(int i = 0; i < n; i++)
		labels_by_size[i] = -1;
	
	int i = 0;
	for(;DPsize[root][i] > k; i++);
	
	recurse_find_dp_size_solution(root, i, chosen);
	
	for(int i = 0; i < L; i++)
		for(int j = 0; j < n; j++)
			if(chosen[i][candidates[i][j]])
				labels_by_size[j] = candidates[i][j];
	
	return get_kmeans_cost(labels_by_size);
}


//////////////////////////////////////////////////////////////////////////////////////////

int main()
{	
	bool synthetic_exp = false; // Run experiments on syntehtic data or real data. Please make sure data is generated.
	int runs = 30;
	
	if(!synthetic_exp)
	{
		int tar_k[25] = {5,8,2,2,2,5,10,7};
		int num_real_data = 8;
		
		const char real_data_name_str[25][205] = {
			"dataset/EUCA/input.txt",
			"dataset/MICE/input.txt",
			"dataset/BIOMED/input.txt",
			"dataset/DIABETES/input.txt",
			"dataset/MAGIC/input.txt",
			"dataset/GESTURE/input.txt",
			"dataset/mnist/input.txt",
			"dataset/zoo/input.txt"
						};
		
		
		for(int i = 0; i < num_real_data; i++)
		{
			cout<<"============================================================"<<endl;
			cout<<"dataset "<<i<<":"<<endl;
			read_data(real_data_name_str[i]);
			
			for(int j = 0; j < runs; j++)
			{
				cout<<"runs "<<j<<":"<<endl;
				
				clock_t time_length1 = clock();
				candidate_clusters();
				build_binary_tree();
				time_length1 = clock() - time_length1;
				
				clock_t time_length2 = clock();
				dp_cost(tar_k[i]);
				cerr<<"dp via cost done"<<endl<<flush;
				double a1 = DPcost[root][tar_k[i]];
				//double a1 = 0;
				time_length2 = clock() - time_length2;
				
				clock_t time_length3 = clock();
				dp_size(1.1);
				cerr<<"dp via size done"<<endl<<flush;
				double a2=find_solution_size(tar_k[i]);
				//double a2 = 0;
				time_length3 = clock() - time_length3;
				
				double seconds_needed1 = (double)time_length1 / CLOCKS_PER_SEC;
				double seconds_needed2 = (double)time_length2 / CLOCKS_PER_SEC;
				double seconds_needed3 = (double)time_length3 / CLOCKS_PER_SEC;
				
				cout<<"seconds: "<<seconds_needed1 + seconds_needed2<<" "<<seconds_needed1 + seconds_needed3<<endl<<flush;
				cout<<"kmeans cost: "<<a1<<" "<<a2<<endl<<flush;
			}
			
			cout<<"============================================================"<<endl;
		}
	}
	
	
	if(synthetic_exp)
	{
		int num_data = 10;
		
		for(int i = 0; i < num_data; i++)
		{
			cout<<"============================================================"<<endl;
			cout<<"dataset "<<i<<":"<<endl;
			read_data(name_str[i]);
			
			for(int j = 0; j < runs; j++)
			{
				cout<<"runs "<<j<<":"<<endl;
				
				clock_t time_length1 = clock();
				candidate_clusters();
				build_binary_tree();
				time_length1 = clock() - time_length1;
				
				clock_t time_length2 = clock();
				dp_cost(target_k[i]);
				cerr<<"dp via cost done"<<endl<<flush;
				double a1 = DPcost[root][target_k[i]];
				//double a1 = 0;
				time_length2 = clock() - time_length2;
				
				clock_t time_length3 = clock();
				dp_size(1.1);
				cerr<<"dp via size done"<<endl<<flush;
				double a2=find_solution_size(target_k[i]);
				//double a2 = 0;
				time_length3 = clock() - time_length3;
				
				double seconds_needed1 = (double)time_length1 / CLOCKS_PER_SEC;
				double seconds_needed2 = (double)time_length2 / CLOCKS_PER_SEC;
				double seconds_needed3 = (double)time_length3 / CLOCKS_PER_SEC;
				
				cout<<"seconds: "<<seconds_needed1 + seconds_needed2<<" "<<seconds_needed1 + seconds_needed3<<endl<<flush;
				cout<<"kmeans cost: "<<a1<<" "<<a2<<endl<<flush;
			}
			cout<<"============================================================"<<endl;
		}
	}
	
	return 0;
}





